import asyncio
from abc import ABC
from datetime import datetime, timedelta
from copy import copy
from typing import Any, List, Union, Optional

from ..objects.interactions import interaction as ext_interaction
from ..utilities import Log


class InteractionResponseHelper(ABC):
    '''Assistive classes to help with responding to an interaction.'''

    _log = Log()

    def __init__(self,
                 target: Union['ext_interaction.Interaction', 'InteractionResponseHelper'],
                 question: str,
                 timeout: timedelta = timedelta(minutes=15),
                 cleanup: bool = False,
                 auto_respond: bool = False,
                 ):
        '''Constructor for abstract Interaction Response Helpers.'''
        self.target = target
        self.question = question
        self.timeout = timeout
        self.cleanup = cleanup
        self.last_interaction: 'ext_interaction.Interaction'
        self.answer: Any = None
        self.auto_respond: bool = auto_respond

        if self.cleanup and self.auto_respond:
            raise ValueError('You cannot set cleanup and auto_respond to \'True\' at the same time.')

        self._id_answer_map: dict = dict()
        self._answered: bool = False
        self._initial_was_response: Optional[bool] = None

    def _ask_base_generation(self) -> Union['ext_interaction.InteractionResponse', 'ext_interaction.InteractionFollowup']:
        '''From the target, generate a base which is appropriate for the interaction.'''
        if isinstance(self.target, InteractionResponseHelper):
            if not self.target._answered:
                raise RuntimeError('Unanswered questions don\'t have interactions yet. You must ask the previous question first.')
            target = self.target.last_interaction
        else:
            target = self.target

        base: Union[ext_interaction.InteractionFollowup, ext_interaction.InteractionResponse]
        if target.can_respond:
            base = target.generate_response(type=target.INTERACTION_RESPONSE_TYPES.CHANNEL_MESSAGE_WITH_SOURCE, ephemeral=True)
            self._initial_was_response = True
        else:
            base = target.generate_followup()
            self._initial_was_response = False
        return base

    async def _wait_for_answer(self):
        start_datetime = datetime.now()

        while datetime.now() < start_datetime + self.timeout:
            await asyncio.sleep(1)
            if self._answered is True:
                break

    async def _call_back(self, client, interaction: 'ext_interaction.Interaction'):
        self._answered = True
        self.last_interaction = interaction
        assert interaction.data is not None
        self.answer = self._id_answer_map[interaction.data.custom_id]
        if self.cleanup:
            response = self.target.generate_followup()
            await response.delete_initial_response()
        if self.auto_respond:
            response = interaction.generate_response(type=interaction.INTERACTION_RESPONSE_TYPES.UPDATE_MESSAGE)
            response.generate(content=f'Question: `{self.question}`, You answered: `{self.answer}`')
            await response.send()

    def generate_response(self, type: 'ext_interaction.enumerations.INTERACTION_RESPONSE_TYPES' = None, ephemeral: bool = True):
        '''Generate a response from the interaction.

        Raises:
            RuntimeError: When cleanup or auto_respond are set to true.
        '''
        if self.cleanup:
            raise RuntimeError('You cannot respond to a message you asked us to cleanup.')
        if self.auto_respond:
            raise RuntimeError('You cannot respond to a message you asked us to auto-respond to.')
        if type is None:
            type = ext_interaction.enumerations.INTERACTION_RESPONSE_TYPES.UPDATE_MESSAGE
        return self.last_interaction.generate_response(type, ephemeral)

    def generate_followup(self):
        '''Generate a followup to the interaction.'''
        if self.cleanup:
            raise RuntimeError('You cannot respond to a message you asked us to cleanup.')
        return self.last_interaction.generate_followup()


class Question(InteractionResponseHelper):
    '''Ask a user a question, respond with the answer.'''

    _log = Log()

    def __init__(self,
                 target: Union['ext_interaction.Interaction', 'InteractionResponseHelper'],
                 question: str,
                 answers: List['str'],
                 timeout: timedelta = timedelta(minutes=15),
                 cleanup: bool = False,
                 auto_respond: bool = False,
                 ):
        '''Prompt user with a question to be answered by one of a list of specific answers.

        Arguments:
            target (Interaction|InteractionResponseHelper): Object to be responded to. Commonly an interaction, but you can chain InteractionResponseHelpers together.
            question (str): Actual text of the question to prompt the user with.
            answers (str): List of valid answers the user can give.
            timeout (timedelta): Time before the response helper gives up.
            cleanup (bool): Should we delete this when an answer is received? Mutually exclusive with `auto_response`.
            auto_respond (bool): Should we give feedback to the user when an answer is received? Mutually exclusive with `clean_up`.
        '''
        super().__init__(target, question, timeout, cleanup, auto_respond)
        self.answers = copy(answers)

        assert len(self.answers) <= 25
        assert type(self.answers) in [list, tuple]

    async def ask(self) -> str:
        '''Ask a question.

        Returns:
            bool: Result from user. Will be `None` if user selected Cancel, or a timeout occurs.
        '''
        assert type(self.target) is ext_interaction.Interaction
        base = self.target.generate_response(type=self.target.INTERACTION_RESPONSE_TYPES.CHANNEL_MESSAGE_WITH_SOURCE)
        base.generate(content=self.question)
        ar = base.add_components()

        for answer in self.answers:
            if hasattr(ar, 'components') and (len(ar.components) == 5):
                ar = base.add_components()
            button = ar.add_button(ar.BUTTON_STYLES.PRIMARY, label=answer, callback=self._call_back)
            self._id_answer_map[button.custom_id] = answer

        await base.send()  # type: ignore
        await self._wait_for_answer()
        return str(self.answer)


class CAPTCHA:
    '''Try to test humans from bots.'''


class Confirmation(InteractionResponseHelper):
    '''Ask a user for a YES/NO/CANCEL response.'''

    _log = Log()

    def __init__(self,
                 target: Union['ext_interaction.Interaction', 'InteractionResponseHelper'],
                 question: str,
                 timeout: timedelta = timedelta(minutes=15),
                 cleanup: bool = False,
                 auto_respond: bool = False,
                 ):
        '''Ask the user to confirm, deny, or cancel.

        Arguments:
            target (Interaction, InteractionResponseHelper): Target of this interaction.
            question (str): Actual text of the question to prompt the user with.
            timeout (timedelta): Time before the response helper gives up.
            cleanup (bool): Should we delete this when an answer is received? Mutually exclusive with `auto_response`.
            auto_respond (bool): Should we give feedback to the user when an answer is received? Mutually exclusive with `clean_up`.
        '''
        super().__init__(target, question, timeout, cleanup, auto_respond)

    async def ask(self) -> bool:
        '''Ask a question.

        Returns:
            bool: Result from user. Will be `None` if user selected Cancel, or a timeout occurs.
        '''
        base = self._ask_base_generation()
        base.generate(content=self.question)
        ar = base.add_components()

        button = ar.add_button(ar.BUTTON_STYLES.SUCCESS, label='Yes', callback=self._call_back)
        self._id_answer_map[button.custom_id] = True

        button = ar.add_button(ar.BUTTON_STYLES.DANGER, label='No', callback=self._call_back)
        self._id_answer_map[button.custom_id] = False

        button = ar.add_button(ar.BUTTON_STYLES.SECONDARY, label='Cancel', callback=self._call_back)
        self._id_answer_map[button.custom_id] = None

        await base.send()
        await self._wait_for_answer()
        return self.answer



import setuptools
from distutils.util import convert_path

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

main_ns = {}
ver_path = convert_path('src/dyscord/version.py')
with open(ver_path) as ver_file:
    exec(ver_file.read(), main_ns)

setuptools.setup(
    name="dyscord",
    version=main_ns['__version__'],
    author="John Murray",
    author_email="jmurrayufo@gmail.com",
    description="A simple implementation of the discord API.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/isbe-house/dyscord",
    project_urls={
        "Bug Tracker": "https://github.com/isbe-house/dyscord/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3.9",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    package_dir={"": "src"},
    packages=setuptools.find_packages(where="src"),
    python_requires=">=3.8",
    license='GPLv3',
    platforms=['any'],
    install_requires=[
        'cachetools>=4.2.4',
        'colour>=0.1.5',
        'emoji>=1.5.0',
        'httpx>=0.19.0',
        'nest_asyncio>=1.5.1',
        'orjson>=3.6.3',
        'requests>=2.26.0',
        'websockets>=10.0',
        'validators>=0.18.2',
    ],
)



'''Provide a alias namespace for common Commanding actions.'''

from ..helper.command_handler import CommandHandler
from ..objects.interactions import Command
from ..objects.interactions.enumerations import COMMAND_TYPE

__all__ = [
    'CommandHandler',
    'Command',
    'COMMAND_TYPE',
]



from typing import Any, Optional

from .base_object import BaseDiscordObject

from . import snowflake


class Role(BaseDiscordObject):
    '''Roles represent a set of permissions attached to a group of users.'''

    id: snowflake.Snowflake = None  # type: ignore
    name: str = None  # type: ignore
    color: int = None  # type: ignore
    hoist: bool = None  # type: ignore
    icon: str = None  # type: ignore
    unicode_emoji: str = None  # type: ignore
    position: int = None  # type: ignore
    permissions: str = None  # type: ignore
    managed: bool = None  # type: ignore
    mentionable: bool = None  # type: ignore
    tags: Optional['RoleTags'] = None  # type: ignore

    def from_dict(self, data: dict) -> 'Role':
        '''Parse a Role from an API compliant dict.'''
        self.id = snowflake.Snowflake(data['id'])
        self.name = data['name']
        self.color = data['color']
        self.hoist = data['hoist']
        self.position = data['position']
        self.permissions = data['permissions']
        self.managed = data['managed']
        self.mentionable = data['mentionable']

        if 'icon' in data:
            self.icon = data['icon']

        if 'unicode_emoji' in data:
            self.unicode_emoji = data['unicode_emoji']

        if 'tags' in data:
            self.tags = RoleTags().from_dict(data['tags'])

        return self


class RoleTags(BaseDiscordObject):
    '''RoleTags.'''

    bot_id: snowflake.Snowflake = None  # type: ignore
    integration_id: snowflake.Snowflake = None  # type: ignore
    premium_subscriber: Any = None  # type: ignore

    def from_dict(self, data: dict) -> 'RoleTags':
        '''Parse a RoleRags from an API compliant dict.'''
        if 'bot_id' in data:
            self.bot_id = data['bot_id']

        if 'integration_id' in data:
            self.integration_id = data['integration_id']

        if 'premium_subscriber' in data:
            self.premium_subscriber = data['premium_subscriber']

        return self



import asyncio
import datetime
from pprint import pprint

import cachetools
import cachetools.keys
import httpx

from ...objects.snowflake import Snowflake

from ...utilities import Log

from ... import objects


class API_V9:
    '''Version 9 of the discord API.'''

    BASE_URL = 'https://discord.com/api/v9'
    TOKEN: str  # Overridden when the API is loaded in DiscordClient._connect
    APPLICATION_ID: str
    USER_AGENT: str = 'Discord Bot ()'

    _lock = asyncio.Lock()
    _log = Log()
    _ttl_cache: dict = cachetools.TTLCache(10_000, ttl=datetime.timedelta(minutes=15), timer=datetime.datetime.now)  # type: ignore

    @classmethod
    def _auth_header(cls):
        return {'Authorization': f'Bot {cls.TOKEN}', 'Content-Type': 'application/json'}

    @classmethod
    async def _handle_rate_limit(cls, response):
        headers = response.headers

        if 'x-ratelimit-reset' not in headers:
            return

        reset_datetime = datetime.datetime.fromtimestamp(float(headers['x-ratelimit-reset']))
        time_until_reset = reset_datetime - datetime.datetime.now()

        if int(headers["x-ratelimit-remaining"]) == 0:
            cls._log.warning(f'Rate limit encountered, waiting for {time_until_reset}.')
            await asyncio.sleep(time_until_reset.total_seconds())

    @classmethod
    async def _invoke_method(cls, method):
        async with cls._lock:
            r = await method
            try:
                r.raise_for_status()
            except Exception:
                cls._log.exception(r.content)
                raise
            await cls._handle_rate_limit(r)
        return r

    # GATEWAY ENDPOINTS

    @classmethod
    async def get_gateway_bot(cls, token: str) -> dict:
        '''Get URL of the gateway for bots.'''
        if not isinstance(token, (str, Snowflake)):
            raise TypeError(f'Got illegal type [{type(token)}] for token.')

        async with cls._lock:
            async with httpx.AsyncClient() as client:
                r = await client.get(
                    f'{cls.BASE_URL}/gateway/bot',
                    headers={'Authorization': f'Bot {token}'},
                )
                r.raise_for_status()
            await cls._handle_rate_limit(r)

        # Add our own settings to URI
        data = r.json()
        data['url'] += '?v=9&encoding=json'

        return data

    @classmethod
    async def get_gateway(cls) -> dict:
        '''Get URL of the gateway.'''
        async with cls._lock:
            async with httpx.AsyncClient() as client:
                r = await client.get(f'{cls.BASE_URL}/gateway')
                r.raise_for_status()
            await cls._handle_rate_limit(r)

        return r.json()

    @classmethod
    async def get_global_application_commands(cls):
        '''Get all global application commands.'''
        url = f'{cls.BASE_URL}/applications/{cls.APPLICATION_ID}/commands'

        async with cls._lock:
            async with httpx.AsyncClient() as client:
                r = await client.get(
                    url,
                    headers=cls._auth_header(),
                )
                r.raise_for_status()
            await cls._handle_rate_limit(r)
        return r.json()

    @classmethod
    async def create_global_application_command(cls, command_structure: dict):
        '''Create or update a global application command.'''
        url = f'{cls.BASE_URL}/applications/{cls.APPLICATION_ID}/commands'

        if not isinstance(command_structure, (dict,)):
            raise TypeError(f'Got illegal type [{type(command_structure)}] for command_structure.')

        async with cls._lock:
            async with httpx.AsyncClient() as client:
                r = await client.post(
                    url,
                    headers=cls._auth_header(),
                    json=command_structure,
                )
                r.raise_for_status()
            await cls._handle_rate_limit(r)
        return r.json()

    @classmethod
    async def get_global_application_command(cls, command_id: 'objects.Snowflake'):
        '''Get a global application command.'''
        url = f'{cls.BASE_URL}/applications/{cls.APPLICATION_ID}/commands/{command_id}'

        if not isinstance(command_id, (str, Snowflake)):
            raise TypeError(f'Got illegal type [{type(command_id)}] for command_id.')

        async with cls._lock:
            async with httpx.AsyncClient() as client:
                r = await client.get(
                    url,
                    headers=cls._auth_header(),
                )
                r.raise_for_status()
            await cls._handle_rate_limit(r)
        return r.json()

    @classmethod
    async def edit_global_application_command(cls,
                                              command_id: 'objects.Snowflake',
                                              command_structure: dict
                                              ):
        '''Edit a global application command.'''
        url = f'{cls.BASE_URL}/applications/{cls.APPLICATION_ID}/commands/{command_id}'

        if not isinstance(command_id, (str, Snowflake)):
            raise TypeError(f'Got illegal type [{type(command_id)}] for command_id.')
        if not isinstance(command_structure, (dict,)):
            raise TypeError(f'Got illegal type [{type(command_structure)}] for command_structure.')

        async with cls._lock:
            async with httpx.AsyncClient() as client:
                r = await client.patch(
                    url,
                    headers=cls._auth_header(),
                    json=command_structure,
                )
                r.raise_for_status()
            await cls._handle_rate_limit(r)
        return r.json()

    @classmethod
    async def delete_global_application_command(cls, command_id: 'objects.Snowflake'):
        '''Delete a global application command.'''
        url = f'{cls.BASE_URL}/applications/{cls.APPLICATION_ID}/commands/{command_id}'

        if not isinstance(command_id, (str, Snowflake)):
            raise TypeError(f'Got illegal type [{type(command_id)}] for command_id.')

        async with cls._lock:
            async with httpx.AsyncClient() as client:
                r = await client.delete(
                    url,
                    headers=cls._auth_header(),
                )
                r.raise_for_status()
            await cls._handle_rate_limit(r)

    @classmethod
    async def bulk_overwrite_global_application_commands(cls, command_id: 'objects.Snowflake'):
        '''Unimplemented.'''
        # PUT/applications/{application.id}/commands
        raise NotImplementedError('TBD')

    @classmethod
    async def get_guild_application_commands(cls, guild_id: 'objects.Snowflake'):
        '''Get all application commands from a specific guild.'''
        url = f'{cls.BASE_URL}/applications/{cls.APPLICATION_ID}/guilds/{guild_id}/commands'

        if not isinstance(guild_id, (str, Snowflake)):
            raise TypeError(f'Got illegal type [{type(guild_id)}] for guild_id.')

        async with cls._lock:
            async with httpx.AsyncClient() as client:
                r = await client.get(
                    url,
                    headers=cls._auth_header(),
                )
                r.raise_for_status()
            await cls._handle_rate_limit(r)
        return r.json()

    @classmethod
    async def create_guild_application_command(cls,
                                               guild_id: 'objects.Snowflake',
                                               command_structure: dict
                                               ) -> dict:
        '''Create a guild appplication command.'''
        url = f'{cls.BASE_URL}/applications/{cls.APPLICATION_ID}/guilds/{guild_id}/commands'

        if not isinstance(guild_id, (str, Snowflake)):
            raise TypeError(f'Got illegal type [{type(guild_id)}] for guild_id.')
        if not isinstance(command_structure, (dict,)):
            raise TypeError(f'Got illegal type [{type(command_structure)}] for command_structure.')

        async with cls._lock:
            async with httpx.AsyncClient() as client:
                r = await client.post(
                    url,
                    headers=cls._auth_header(),
                    json=command_structure,
                )
                try:
                    r.raise_for_status()
                except Exception:
                    pprint(command_structure)
                    pprint(r.json())
                    raise
            await cls._handle_rate_limit(r)
        return r.json()

    @classmethod
    async def get_guild_application_command(cls,
                                            guild_id: 'objects.Snowflake',
                                            command_id: 'objects.Snowflake',
                                            ) -> dict:
        '''Get a specific guild application command.'''
        url = f'{cls.BASE_URL}/applications/{cls.APPLICATION_ID}/guilds/{guild_id}/commands/{command_id}'

        if not isinstance(guild_id, (str, Snowflake)):
            raise TypeError(f'Got illegal type [{type(guild_id)}] for guild_id.')
        if not isinstance(command_id, (str, Snowflake)):
            raise TypeError(f'Got illegal type [{type(command_id)}] for command_id.')

        async with cls._lock:
            async with httpx.AsyncClient() as client:
                r = await client.get(
                    url,
                    headers=cls._auth_header(),
                )
                r.raise_for_status()
            await cls._handle_rate_limit(r)
        return r.json()

    @classmethod
    async def edit_guild_application_command(cls,
                                             guild_id: 'objects.Snowflake',
                                             command_id: 'objects.Snowflake',
                                             command_structure: dict,
                                             ):
        '''Edit a guild application command.'''
        # PATCH /applications/{application.id}/guilds/{guild.id}/commands/{command.id}
        raise NotImplementedError('TBD')

    @classmethod
    async def delete_guild_application_command(cls,
                                               guild_id: 'objects.Snowflake',
                                               command_id: 'objects.Snowflake',
                                               ) -> None:
        '''Delete an application command.'''
        url = f'{cls.BASE_URL}/applications/{cls.APPLICATION_ID}/guilds/{guild_id}/commands/{command_id}'

        if not isinstance(guild_id, (str, Snowflake)):
            raise TypeError(f'Got illegal type [{type(guild_id)}] for guild_id.')
        if not isinstance(command_id, (str, Snowflake)):
            raise TypeError(f'Got illegal type [{type(command_id)}] for command_id.')

        async with cls._lock:
            async with httpx.AsyncClient() as client:
                r = await client.delete(
                    url,
                    headers=cls._auth_header(),
                )
                try:
                    r.raise_for_status()
                except Exception:
                    pprint(r.json())
                    raise
            await cls._handle_rate_limit(r)

    @classmethod
    async def bulk_overwrite_guild_application_command(cls,
                                                       guild_id: 'objects.Snowflake',
                                                       command_id: 'objects.Snowflake',
                                                       command_structure: dict,
                                                       ):
        '''TODO: Copy from api docs.'''
        # PUT /applications/{application.id}/guilds/{guild.id}/commands/{command.id}
        raise NotImplementedError('TBD')

    # Interaction Methods
    '''
    '''

    @classmethod
    async def create_interaction_response(cls,
                                          interaction_id: 'objects.Snowflake',
                                          interaction_token: str,
                                          data_structure: dict,
                                          ) -> None:
        '''TODO: Copy from api docs.'''
        url = f'{cls.BASE_URL}/interactions/{interaction_id}/{interaction_token}/callback'

        if not isinstance(interaction_id, (str, Snowflake)):
            raise TypeError(f'Got illegal type [{type(interaction_id)}] for interaction_id.')
        if not isinstance(interaction_token, (str, )):
            raise TypeError(f'Got illegal type [{type(interaction_token)}] for interaction_token.')
        if not isinstance(data_structure, (dict, )):
            raise TypeError(f'Got illegal type [{type(data_structure)}] for data_structure.')

        async with cls._lock:
            async with httpx.AsyncClient() as client:
                r = await client.post(
                    url,
                    json=data_structure
                )
                try:
                    r.raise_for_status()
                except Exception:
                    print(r.content)
                    raise
            await cls._handle_rate_limit(r)

    @classmethod
    async def get_original_interaction_response(cls,
                                                interaction_token: str,
                                                ) -> dict:
        '''TODO: Copy from api docs.'''
        url = f'{cls.BASE_URL}/webhooks/{cls.APPLICATION_ID}/{interaction_token}/messages/@origional'

        if not isinstance(interaction_token, (str, )):
            raise TypeError(f'Got illegal type [{type(interaction_token)}] for interaction_token.')

        async with cls._lock:
            async with httpx.AsyncClient() as client:
                r = await client.get(
                    url,
                )
                r.raise_for_status()
            await cls._handle_rate_limit(r)
        return r.json()

    @classmethod
    async def edit_original_interaction_response(cls,
                                                 interaction_token: str,
                                                 data_structure: dict,
                                                 ) -> None:
        '''Edit Original Interaction Response.

        PATCH/webhooks/{application.id}/{interaction.token}/messages/@original

        Edits the initial Interaction response. Functions the same as Edit Webhook Message.
        '''
        url = f'{cls.BASE_URL}/webhooks/{cls.APPLICATION_ID}/{interaction_token}/messages/@original'

        if not isinstance(interaction_token, (str, )):
            raise TypeError(f'Got illegal type [{type(interaction_token)}] for interaction_token.')
        if not isinstance(data_structure, (dict, )):
            raise TypeError(f'Got illegal type [{type(data_structure)}] for data_structure.')

        async with cls._lock:
            async with httpx.AsyncClient() as client:
                r = await client.patch(
                    url,
                    json=data_structure
                )
                r.raise_for_status()
            await cls._handle_rate_limit(r)

    @classmethod
    async def delete_original_interaction_response(cls,
                                                   interaction_token: str,
                                                   ) -> None:
        '''Delete Original Interaction Response.

        DELETE/webhooks/{application.id}/{interaction.token}/messages/@original

        Deletes the initial Interaction response. Returns 204 on success.
        '''
        url = f'{cls.BASE_URL}/webhooks/{cls.APPLICATION_ID}/{interaction_token}/messages/@original'

        async with cls._lock:
            async with httpx.AsyncClient() as client:
                r = await client.delete(
                    url,
                )
                r.raise_for_status()
            await cls._handle_rate_limit(r)

        if not isinstance(interaction_token, (str, )):
            raise TypeError(f'Got illegal type [{type(interaction_token)}] for interaction_token.')

    @classmethod
    async def create_followup_message(cls,
                                      interaction_token: str,
                                      data_structure: dict,
                                      ) -> dict:
        '''Create Followup Message.

        POST/webhooks/{application.id}/{interaction.token}

        Create a followup message for an Interaction. Functions the same as Execute Webhook, but wait is always true,
        and flags can be set to 64 in the body to send an ephemeral message. The thread_id query parameter is not required
        (and is furthermore ignored) when using this endpoint for interaction followups.
        '''
        url = f'{cls.BASE_URL}/webhooks/{cls.APPLICATION_ID}/{interaction_token}'

        if not isinstance(interaction_token, (str, )):
            raise TypeError(f'Got illegal type [{type(interaction_token)}] for interaction_token.')
        if not isinstance(data_structure, (dict,)):
            raise TypeError(f'Got illegal type [{type(data_structure)}] for data_structure.')

        async with cls._lock:
            async with httpx.AsyncClient() as client:
                r = await client.post(
                    url,
                    json=data_structure
                )
                try:
                    r.raise_for_status()
                except Exception:
                    print(r)
                    raise
            await cls._handle_rate_limit(r)
        return r.json()

    @classmethod
    async def get_followup_message(cls,
                                   interaction_token: str,
                                   message_id: 'Snowflake',
                                   ) -> dict:
        '''Get Followup Message.

        GET/webhooks/{application.id}/{interaction.token}/messages/{message.id}

        Returns a followup message for an Interaction. Functions the same as Get Webhook Message. Does not support ephemeral followups.
        '''
        url = f'{cls.BASE_URL}/webhooks/{cls.APPLICATION_ID}/{interaction_token}/messages/{message_id}'

        if not isinstance(interaction_token, (str, )):
            raise TypeError(f'Got illegal type [{type(interaction_token)}] for interaction_token.')
        if not isinstance(message_id, (str, Snowflake)):
            raise TypeError(f'Got illegal type [{type(message_id)}] for message_id.')

        async with cls._lock:
            async with httpx.AsyncClient() as client:
                r = await client.get(
                    url
                )
                try:
                    r.raise_for_status()
                except Exception:
                    print(r)
                    raise
            await cls._handle_rate_limit(r)
        return r.json()

    @classmethod
    async def edit_followup_message(cls,
                                    interaction_token: str,
                                    message_id: 'Snowflake',
                                    data_structure: dict,
                                    ) -> dict:
        '''Edit Followup Message.

        PATCH/webhooks/{application.id}/{interaction.token}/messages/{message.id}

        Edits a followup message for an Interaction. Functions the same as Edit Webhook Message. Does not support ephemeral followups.
        '''
        url = f'{cls.BASE_URL}/webhooks/{cls.APPLICATION_ID}/{interaction_token}/messages/{message_id}'

        if not isinstance(interaction_token, (str, )):
            raise TypeError(f'Got illegal type [{type(interaction_token)}] for interaction_token.')
        if not isinstance(message_id, (str, Snowflake)):
            raise TypeError(f'Got illegal type [{type(message_id)}] for message_id.')
        if not isinstance(data_structure, (dict, )):
            raise TypeError(f'Got illegal type [{type(data_structure)}] for data_structure.')

        async with cls._lock:
            async with httpx.AsyncClient() as client:
                r = await client.patch(
                    url,
                    json=data_structure
                )
                r.raise_for_status()
            await cls._handle_rate_limit(r)
        return r.json()

    @classmethod
    async def delete_followup_message(cls,
                                      interaction_token: str,
                                      message_id: 'Snowflake',
                                      ) -> None:
        '''Delete Followup Message.

        DELETE/webhooks/{application.id}/{interaction.token}/messages/{message.id}

        Deletes a followup message for an Interaction. Returns 204 on success. Does not support ephemeral followups.
        '''
        url = f'{cls.BASE_URL}/webhooks/{cls.APPLICATION_ID}/{interaction_token}/messages/{message_id}'

        if not isinstance(interaction_token, (str,)):
            raise TypeError(f'Got illegal type [{type(interaction_token)}] for interaction_token.')
        if not isinstance(message_id, (str, Snowflake)):
            raise TypeError(f'Got illegal type [{type(message_id)}] for message_id.')

        async with cls._lock:
            async with httpx.AsyncClient() as client:
                r = await client.delete(
                    url
                )
                r.raise_for_status()
            await cls._handle_rate_limit(r)

    '''
    TODO: Implement the following API endpoints.

    Get Guild Application Command Permissions
        GET/applications/{application.id}/guilds/{guild.id}/commands/permissions

    Get Application Command Permissions
        GET/applications/{application.id}/guilds/{guild.id}/commands/{command.id}/permissions

    Edit Application Command Permissions
        PUT/applications/{application.id}/guilds/{guild.id}/commands/{command.id}/permissions

    Batch Edit Application Command Permissions
        PUT/applications/{application.id}/guilds/{guild.id}/commands/permissions
    '''

    # Channel methods

    @classmethod
    async def get_channel(cls, channel_id: 'objects.Snowflake') -> dict:
        '''Get channel by ID.'''
        url = f'{cls.BASE_URL}/channels/{channel_id}'

        if not isinstance(channel_id, (str, Snowflake)):
            raise TypeError(f'Got illegal type [{type(channel_id)}] for channel_id.')

        try:
            return cls._ttl_cache[cachetools.keys.hashkey('get_channel', channel_id)]
        except KeyError:
            pass

        async with cls._lock:
            async with httpx.AsyncClient() as client:
                r = await client.get(
                    url,
                    headers=cls._auth_header(),
                )
                r.raise_for_status()
            await cls._handle_rate_limit(r)
        cls._ttl_cache[cachetools.keys.hashkey('get_channel', channel_id)] = r.json()
        return r.json()

    @classmethod
    async def create_message(cls,
                             channel_id: 'objects.Snowflake',
                             message_payload: dict,
                             ):
        '''TODO: Copy from api docs.'''
        url = f'{cls.BASE_URL}/channels/{channel_id}/messages'

        if not isinstance(channel_id, (str, Snowflake)):
            raise TypeError(f'Got illegal type [{type(channel_id)}] for channel_id.')
        if not isinstance(message_payload, (dict,)):
            raise TypeError(f'Got illegal type [{type(message_payload)}] for message_payload.')

        async with cls._lock:
            async with httpx.AsyncClient() as client:
                r = await client.post(
                    url,
                    headers=cls._auth_header(),
                    json=message_payload,
                )
                try:
                    r.raise_for_status()
                except Exception:
                    cls._log.exception(r.content)
                    raise
            await cls._handle_rate_limit(r)
        return r.json()

    # Guild methods

    @classmethod
    async def get_guild(cls, guild_id: 'objects.Snowflake') -> dict:
        '''Get guilds.

        TODO: Document this.
        '''
        url = f'{cls.BASE_URL}/guilds/{guild_id}'

        if not isinstance(guild_id, (str, Snowflake)):
            raise TypeError(f'Got illegal type [{type(guild_id)}] for guild_id.')

        try:
            return cls._ttl_cache[cachetools.keys.hashkey('get_guild', guild_id)]
        except KeyError:
            pass

        async with cls._lock:
            async with httpx.AsyncClient() as client:
                r = await client.get(
                    url,
                    headers=cls._auth_header(),
                )
                try:
                    r.raise_for_status()
                except Exception:
                    cls._log.exception(r.content)
                    raise
            await cls._handle_rate_limit(r)
        cls._ttl_cache[cachetools.keys.hashkey('get_guild', guild_id)] = r.json()
        return r.json()

    @classmethod
    async def get_guild_roles(cls, guild_id: 'Snowflake') -> dict:
        '''Get Guild Roles.

        GET/guilds/{guild.id}/roles

        Returns a list of role objects for the guild.
        '''
        url = f'{cls.BASE_URL}/guilds/{guild_id}/roles'

        if not isinstance(guild_id, (str, Snowflake)):
            raise TypeError(f'Got illegal type [{type(guild_id)}] for guild_id.')

        try:
            return cls._ttl_cache[cachetools.keys.hashkey('get_guild_roles', guild_id)]
        except KeyError:
            pass

        async with httpx.AsyncClient() as client:
            method = client.get(url, headers=cls._auth_header())
            r = await cls._invoke_method(method)
        cls._ttl_cache[cachetools.keys.hashkey('get_guild_roles', guild_id)] = r.json()
        return r.json()

    # User methods

    @classmethod
    async def get_current_user(cls) -> dict:
        '''Get Current User.

        GET/users/@me

        Returns the user object of the requester's account. For OAuth2, this requires the identify scope, which will return the object without an email,
        and optionally the email scope, which returns the object with an email.
        '''
        url = f'{cls.BASE_URL}/users/@me'

        async with httpx.AsyncClient() as client:
            method = client.get(url, headers=cls._auth_header())
            r = await cls._invoke_method(method)
        return r.json()

    @classmethod
    async def get_user(cls, user_id: 'Snowflake') -> dict:
        '''Get User.

        GET/users/{user.id}

        Returns a user object for a given user ID.
        '''
        url = f'{cls.BASE_URL}/users/{user_id}'

        if not isinstance(user_id, (str, Snowflake)):
            raise TypeError(f'Got illegal type [{type(user_id)}] for user_id.')

        try:
            return cls._ttl_cache[cachetools.keys.hashkey('get_user', user_id)]
        except KeyError:
            pass

        async with httpx.AsyncClient() as client:
            method = client.get(url, headers=cls._auth_header())
            r = await cls._invoke_method(method)
        cls._ttl_cache[cachetools.keys.hashkey('get_user', user_id)] = r.json()
        return r.json()

    @classmethod
    async def create_dm(cls, recipient_id: 'Snowflake') -> dict:
        '''Create DM.

        POST/users/@me/channels

        Create a new DM channel with a user. Returns a DM channel object.
        '''
        url = f'{cls.BASE_URL}/users/@me/channels'

        if not isinstance(recipient_id, (str, Snowflake)):
            raise TypeError(f'Got illegal type [{type(recipient_id)}] for recipient_id.')

        async with httpx.AsyncClient() as client:
            method = client.post(url,
                                 json={'recipient_id': recipient_id},
                                 headers=cls._auth_header(),
                                 )
            r = await cls._invoke_method(method)
        return r.json()

    '''
    Modify Current User
        PATCH/users/@me
        Modify the requester's user account settings. Returns a user object on success.

    Get Current User Guilds
        GET/users/@me/guilds
        Returns a list of partial guild objects the current user is a member of. Requires the guilds OAuth2 scope.

    Leave Guild
        DELETE/users/@me/guilds/{guild.id}
        Leave a guild. Returns a 204 empty response on success.

    Get User Connections
        GET/users/@me/connections
        Returns a list of connection objects. Requires the connections OAuth2 scope.
    '''  # noqa: E501

    # Webhook methods

    '''
    Create Webhook
        POST/channels/{channel.id}/webhooks

    Get Channel Webhooks
        GET/channels/{channel.id}/webhooks

    Get Guild Webhooks
        GET/guilds/{guild.id}/webhooks

    Get Webhook
        GET/webhooks/{webhook.id}

    Get Webhook with Token
        GET/webhooks/{webhook.id}/{webhook.token}

    Modify Webhook
        PATCH/webhooks/{webhook.id}

    Modify Webhook with Token
        PATCH/webhooks/{webhook.id}/{webhook.token}

    Delete Webhook
        DELETE/webhooks/{webhook.id}

    Delete Webhook with Token
        DELETE/webhooks/{webhook.id}/{webhook.token}

    Execute Webhook
        POST/webhooks/{webhook.id}/{webhook.token}

    Execute Slack-Compatible Webhook
        POST/webhooks/{webhook.id}/{webhook.token}/slack

    Execute GitHub-Compatible Webhook
        POST/webhooks/{webhook.id}/{webhook.token}/github

    Get Webhook Message
        GET/webhooks/{webhook.id}/{webhook.token}/messages/{message.id}

    Edit Webhook Message
        PATCH/webhooks/{webhook.id}/{webhook.token}/messages/{message.id}

    Delete Webhook Message
        DELETE/webhooks/{webhook.id}/{webhook.token}/messages/{message.id}
    '''  # noqa: E501



import enum
from typing import Optional, List

from .base_object import BaseDiscordObject

from . import snowflake, channel as ext_channel


class Guild(BaseDiscordObject):
    '''Commonly know as a server.'''

    class VerificationLevels(enum.IntEnum):
        '''Level of verification required by the guild.

        TODO: Move this to enumerations.
        '''
        NONE = 0  # unrestricted
        LOW = 1  # must have verified email on account
        MEDIUM = 2  # must be registered on Discord for longer than 5 minutes
        HIGH = 3  # must be a member of the server for longer than 10 minutes
        VERY_HIGH = 4  # must have a verified phone number

    class NSFWLevels(enum.IntEnum):
        '''NSFW level allowed in guild.

        TODO: Move this to enumerations.
        '''
        DEFAULT = 0
        EXPLICIT = 1
        SAFE = 2
        AGE_RESTRICTED = 3

    class PremiumTiers(enum.IntEnum):
        '''Premium tier currently unlocked by guild.

        TODO: Move this to enumerations.
        '''
        NONE = 0  # guild has not unlocked any Server Boost perks
        TIER_1 = 1  # guild has unlocked Server Boost level 1 perks
        TIER_2 = 2  # guild has unlocked Server Boost level 2 perks
        TIER_3 = 3  # guild has unlocked Server Boost level 3 perks

    class FeatureFlags(enum.Flag):
        '''Features of guild.

        TODO: Move this to enumerations.
        '''
        ANIMATED_ICON = enum.auto()  # guild has access to set an animated guild icon
        BANNER = enum.auto()  # guild has access to set a guild banner image
        COMMERCE = enum.auto()  # guild has access to use commerce features (i.e. create store channels)
        COMMUNITY = enum.auto()  # guild can enable welcome screen, Membership Screening, stage channels and discovery, and receives community updates
        DISCOVERABLE = enum.auto()  # guild is able to be discovered in the directory
        FEATURABLE = enum.auto()  # guild is able to be featured in the directory
        INVITE_SPLASH = enum.auto()  # guild has access to set an invite splash background
        MEMBER_VERIFICATION_GATE_ENABLED = enum.auto()  # guild has enabled Membership Screening
        NEWS = enum.auto()  # guild has access to create news channels
        PARTNERED = enum.auto()  # guild is partnered
        PREVIEW_ENABLED = enum.auto()  # guild can be previewed before joining via Membership Screening or the directory
        VANITY_URL = enum.auto()  # guild has access to set a vanity URL
        VERIFIED = enum.auto()  # guild is verified
        VIP_REGIONS = enum.auto()  # guild has access to set 384kbps bitrate in voice (previously VIP voice servers)
        WELCOME_SCREEN_ENABLED = enum.auto()  # guild has enabled the welcome screen
        TICKETED_EVENTS_ENABLED = enum.auto()  # guild has enabled ticketed events
        MONETIZATION_ENABLED = enum.auto()  # guild has enabled monetization
        MORE_STICKERS = enum.auto()  # guild has increased custom sticker slots
        THREE_DAY_THREAD_ARCHIVE = enum.auto()  # guild has access to the three day archive time for threads
        SEVEN_DAY_THREAD_ARCHIVE = enum.auto()  # guild has access to the seven day archive time for threads
        PRIVATE_THREADS = enum.auto()  # guild has access to create private threads

    id: snowflake.Snowflake = None  # type: ignore
    name: str = None  # type: ignore
    icon: str = None  # type: ignore
    icon_hash: Optional[str] = None  # type: ignore
    splash: Optional[str] = None  # type: ignore
    discovery_splash: Optional[str] = None  # type: ignore

    am_owner: bool = None  # type: ignore # boolean true if the user is the owner of the guild
    owner_id: Optional['snowflake.Snowflake'] = None  # type: ignore # snowflake id of owner
    permissions: type = None  # type: ignore # string total permissions for the user in the guild (excludes overwrites)
    region: type = None  # type: ignore # string voice region id for the guild (deprecated)
    afk_channel_id: type = None  # type: ignore # snowflake id of afk channel
    afk_timeout: type = None  # type: ignore # integer afk timeout in seconds
    widget_enabled: type = None  # type: ignore # boolean true if the server widget is enabled
    widget_channel_id: type = None  # type: ignore # snowflake the channel id that the widget will generate an invite to, or null if set to no invite
    verification_level: type = None  # type: ignore # integer verification level required for the guild
    default_message_notifications: type = None  # type: ignore # integer default message notifications level
    explicit_content_filter: type = None  # type: ignore # integer explicit content filter level
    roles: type = None  # type: ignore # array of role objects roles in the guild
    emojis: type = None  # type: ignore # array of emoji objects custom guild emojis
    features: type = None  # type: ignore # array of guild feature strings enabled guild features
    mfa_level: type = None  # type: ignore # integer required MFA level for the guild
    application_id: type = None  # type: ignore # snowflake application id of the guild creator if it is bot-created
    system_channel_id: type = None  # type: ignore # snowflake the id of the channel where guild notices such as welcome messages and boost events are posted
    system_channel_flags: type = None  # type: ignore # integer system channel flags
    rules_channel_id: type = None  # type: ignore # snowflake the id of the channel where Community guilds can display rules and/or guidelines
    joined_at: type = None  # type: ignore # * ISO8601 timestamp when this guild was joined at
    large: type = None  # type: ignore # * boolean true if this is considered a large guild
    unavailable: type = None  # type: ignore # * boolean true if this guild is unavailable due to an outage
    member_count: type = None  # type: ignore # * integer total number of members in this guild
    voice_states: type = None  # type: ignore # * array of partial voice state objects states of members currently in voice channels; lacks the guild_id key
    members: type = None  # type: ignore # * array of guild member objects users in the guild
    channels: List['ext_channel.Channel'] = None  # type: ignore # * array of channel objects channels in the guild
    threads: type = None  # type: ignore # * array of channel objects all active threads in the guild that current user has permission to view
    presences: type = None  # type: ignore # * array of partial presence update objects presences of the members in the guild,
    #  will only include non-offline members if the size is greater than large threshold
    max_presences: type = None  # type: ignore # integer the maximum number of presences for the guild (null is always returned, apart from the largest of guilds)
    max_members: type = None  # type: ignore # integer the maximum number of members for the guild
    vanity_url_code: type = None  # type: ignore # string the vanity url code for the guild
    description: type = None  # type: ignore # string the description of a Community guild
    banner: type = None  # type: ignore # string banner hash
    premium_tier: type = None  # type: ignore # integer premium tier (Server Boost level)
    premium_subscription_count: type = None  # type: ignore # integer the number of boosts this guild currently has
    preferred_locale: type = None  # type: ignore # string the preferred locale of a Community guild; used in server discovery and notices from Discord; defaults to "en-US"
    public_updates_channel_id: type = None  # type: ignore # snowflake the id of the channel where admins and moderators of Community guilds receive notices from Discord
    max_video_channel_users: type = None  # type: ignore # integer the maximum amount of users in a video channel
    approximate_member_count: type = None  # type: ignore # integer approximate number of members in this guild, returned from the GET /guilds/<id> endpoint when with_counts is true
    approximate_presence_count: type = None  # type: ignore # integer approximate number of non-offline members in this guild, returned from the GET /guilds/<id> endpoint when with_counts is true
    welcome_screen: type = None  # type: ignore # welcome screen object the welcome screen of a Community guild, shown to new members, returned in an Invite's guild object
    nsfw_level: type = None  # type: ignore # integer guild NSFW level
    stage_instances: type = None  # type: ignore # * array of stage instance objects Stage instances in the guild
    stickers: type = None  # type: ignore # array of sticker objects custom guild stickers

    def __str__(self):
        '''Return string representation.'''
        fields = []

        if self.id is not None:
            fields.append(f'id={self.id}')

        if self.name is not None:
            fields.append(f'name={self.name}')

        return f'Guild({", ".join(fields)})'

    def __repr__(self):
        '''Return string representation.'''
        return self.__str__()

    def from_dict(self, data: dict) -> 'Guild':
        '''Parse a Guild from an API compliant dict.'''
        self.id = snowflake.Snowflake(data.get('id'))
        self.name = data['name']
        self.icon = data['icon']
        self.icon_hash = data.get('icon_hash', None)
        self.splash = data.get('splash', None)
        self.afk_channel_id = data.get('afk_channel_id', None)
        self.verification_level = data.get('verification_level', None)
        self.discovery_splash = data.get('discovery_splash', None)
        if 'owner' in data:
            self.am_owner = data['owner']
        if 'owner_id' in data:
            self.owner_id = snowflake.Snowflake(data['owner_id'])

        self.approximate_member_count = data.get('approximate_member_count', None)
        self.approximate_presence_count = data.get('approximate_presence_count', None)

        self.channels = []
        # Process in channels
        if 'channels' in data:
            for channel_dict in data['channels']:
                # Generate channel
                new_channel = ext_channel.ChannelImporter.from_dict(channel_dict, self)
                self.channels.append(new_channel)

        return self




import enum


class CHANNEL_TYPE(enum.IntEnum):
    '''Types of channels.'''
    GUILD_TEXT = 0  # a text channel within a server
    DM = 1  # a direct message between users
    GUILD_VOICE = 2  # a voice channel within a server
    GROUP_DM = 3  # a direct message between multiple users
    GUILD_CATEGORY = 4  # an organizational category that contains up to 50 channels
    GUILD_NEWS = 5  # a channel that users can follow and crosspost into their own server
    GUILD_STORE = 6  # a channel in which game developers can sell their game on Discord
    GUILD_NEWS_THREAD = 10  # a temporary sub-channel within a GUILD_NEWS channel
    GUILD_PUBLIC_THREAD = 11  # a temporary sub-channel within a GUILD_TEXT channel
    GUILD_PRIVATE_THREAD = 12  # a temporary sub-channel within a GUILD_TEXT channel that is only viewable by those invited and those with the MANAGE_THREADS permission
    GUILD_STAGE_VOICE = 13  # a voice channel for hosting events with an audience


class COMMAND_TYPE(enum.IntEnum):
    '''Type of command.

    Attributes:
        CHAT_INPUT (int): Slash commands; a text-based command that shows up when a user types.
        USER (int): A UI-based command that shows up when you right click or tap on a user.
        MESSAGE (int): A UI-based command that shows up when you right click or tap on a message.
    '''
    CHAT_INPUT = 1  # Slash commands; a text-based command that shows up when a user types
    USER = 2  # A UI-based command that shows up when you right click or tap on a user
    MESSAGE = 3  # A UI-based command that shows up when you right click or tap on a message


class COMMAND_OPTION(enum.IntEnum):
    '''Valid types or groupings of command options.'''
    SUB_COMMAND = 1
    SUB_COMMAND_GROUP = 2
    STRING = 3
    INTEGER = 4  # Any integer between -2^53 and 2^53
    BOOLEAN = 5
    USER = 6
    CHANNEL = 7  # Includes all channel types + categories
    ROLE = 8
    MENTIONABLE = 9  # Includes users and roles
    NUMBER = 10  # Any double between -2^53 and 2^53


class COMPONENT_TYPES(enum.IntEnum):
    '''Types of Components.'''
    ACTION_ROW = 1
    BUTTON = 2
    SELECT_MENU = 3


class BUTTON_STYLES(enum.IntEnum):
    '''Styles of Button objects.'''
    PRIMARY = 1
    SECONDARY = 2
    SUCCESS = 3
    DANGER = 4
    LINK = 5


class INTERACTION_TYPES(enum.IntEnum):
    '''Types of interactions.'''
    PING = 1
    APPLICATION_COMMAND = 2
    MESSAGE_COMPONENT = 3


class INTERACTION_RESPONSE_TYPES(enum.IntEnum):
    '''Types of interaction responses.

    Attributes:
        PONG: ACK a Ping
        CHANNEL_MESSAGE_WITH_SOURCE: Respond to an interaction with a message.
        DEFERRED_CHANNEL_MESSAGE_WITH_SOURCE: ACK an interaction and edit a response later, the user sees a loading state.
        DEFERRED_UPDATE_MESSAGE: Only for components (Not slash commands), ACK an interaction and edit the original message later; the user does not see a loading state.
        UPDATE_MESSAGE: Only for components (Not slash commands), edit the message the component was attached to

    '''
    PONG = 1  # ACK a Ping
    CHANNEL_MESSAGE_WITH_SOURCE = 4  # respond to an interaction with a message
    DEFERRED_CHANNEL_MESSAGE_WITH_SOURCE = 5  # ACK an interaction and edit a response later, the user sees a loading state
    DEFERRED_UPDATE_MESSAGE = 6  # for components, ACK an interaction and edit the original message later; the user does not see a loading state
    UPDATE_MESSAGE = 7  # for components, edit the message the component was attached to


class INTERACTION_CALLBACK_FLAGS(enum.IntFlag):
    '''Flags for callback methods.'''
    EPHEMERAL = (1 << 6)  # only the user receiving the message can see it



import asyncio
import datetime
import enum
from typing import Union, Optional, List, Dict

from .base_object import BaseDiscordObject
from . import snowflake, enumerations
from . import user as ext_user,\
    channel as ext_channel,\
    embed as ext_embed,\
    guild as ext_guild,\
    role as ext_role
from ..utilities import log
from .interactions import components as ext_components

from ..client import api


class Message(BaseDiscordObject, ext_components.ComponentAdder, ext_embed.EmbedAdder):
    '''Message containing infomation about it's content, origin, authors, etc.

    Attributes:
        id (Snowflake): Unique ID of the message.
        channel_id (Snowflake): Unique ID of the channel the message came from.
        guild_id (Snowflake|None): Unique ID of the channel the message came from. Defaults to None.
        guild (Guild|None): Actual guild object item is from. This will autopopulate from the API if it is not present.
        author (User): The author of the message.
        member (Member): If the message was in a guild, this will be the Member object of the author.
        content (str): The actual content of the message.
        timestamp (datetime): Date and time the message was posted.
        edited_timestamp (datetime|None): If the message has been edited, this is the date and time it was.
        tts (bool): Text to speech flag.
        mention_everyone (bool): Message had an @everyone in it.
        mentions ([User]): List of mentioned User.
        mention_roles ([Role]): List of mentioned Role.
        mention_channels ([Channel]): List of mentioned Channel. Note that each channel will be the correct type.
        embeds ([Embed]): List of rich Embed objects.
        nonce (str|None): Nonce used to send the message.
        pinned (bool): True if message is pinned in the channel.
        webhook_id (Snowflake|None): ID of the webhook used to send this message.
        type (MESSAGE_TYPE): Enumeration of the type of message.
        application_id (Snowflake|None): ID of the application that send the message.
        message_reference (MessageReference|None): data showing the source of a crosspost, channel follow add, pin, or reply message
        flags (int): Bitfield of message flags.
        referenced_message (Message|None): Message object refereed to in message_reference.
        thread (Channel): Thread the message is a part of.
        components ([Component]): List of components in the message if it is an interaction.
    '''

    _log = log.Log()

    MESSAGE_TYPE = enumerations.MESSAGE_TYPE

    id: 'snowflake.Snowflake' = None  # type: ignore
    channel_id: 'snowflake.Snowflake' = None  # type: ignore
    guild_id: Optional['snowflake.Snowflake'] = None  # type: ignore
    author: 'ext_user.User' = None  # type: ignore
    member: Optional['ext_user.Member'] = None  # type: ignore
    content: str = None  # type: ignore
    timestamp: datetime.datetime = None  # type: ignore
    edited_timestamp: Optional[datetime.datetime] = None  # type: ignore
    tts: bool = None  # type: ignore
    mention_everyone: bool = None  # type: ignore
    mentions: List['ext_user.User'] = None  # type: ignore
    mention_roles: List['ext_role.Role'] = None  # type: ignore
    mention_channels: List['ext_channel.Channel'] = None  # type: ignore
    attachments: 'List' = None  # type: ignore
    embeds: Optional[List[ext_embed.Embed]] = None  # type: ignore
    # reactions: 'List[reaction.Reaction]'  = None  # type: ignore
    nonce: Optional[Union[int, str]] = None  # type: ignore
    pinned: bool = None  # type: ignore
    webhook_id: Optional['snowflake.Snowflake'] = None  # type: ignore
    type: 'enumerations.MESSAGE_TYPE' = None  # type: ignore
    # activity = None  # Activity TBD = None  # type: ignore
    # application: application.Application = None  # Application TBD = None  # type: ignore
    application_id: 'Optional[snowflake.Snowflake]' = None  # type: ignore
    message_reference: 'Optional[MessageReference]' = None  # type: ignore
    flags: Optional[int]  # type: ignore
    referenced_message: 'Optional[Message]' = None  # type: ignore
    # interaction = None  # type: ignore
    thread: 'ext_channel.Channel' = None  # type: ignore
    components: List['ext_components.Component'] = None  # type: ignore
    # sticker_items: List[Sticker] = []  # type: ignore

    def __init__(self,
                 content: str = None
                 ):
        '''Init.'''
        if content is not None:
            self.content = content

    @property
    def channel(self):
        '''Attempt to grab channel from the API.'''
        if self.channel_id is None:
            return None
        loop = asyncio.get_event_loop()
        channel_dict = loop.run_until_complete(api.API.get_channel(self.channel_id))
        channel = ext_channel.ChannelImporter().from_dict(channel_dict)
        self._log.info('Got channel from the API.')
        return channel

    @property
    def guild(self):
        '''Attempt to grab guild from the API.'''
        if self.guild_id is None:
            return None
        loop = asyncio.get_event_loop()
        guild_dict = loop.run_until_complete(api.API.get_guild(self.guild_id))
        guild = ext_guild.Guild().from_dict(guild_dict)
        self._log.info('Got guild from the API.')
        return guild

    def from_dict(self, data: dict) -> 'Message':  # noqa: C901
        '''Parse a Message from an API compliant dict.'''
        # Mandatory Fields
        self.id = snowflake.Snowflake(data['id'])

        if 'attachments' in data:
            self.attachments = list()
            for attachment in data['attachments']:
                self.attachments.append(attachment)
        if 'author' in data:
            self.author = ext_user.User().from_dict(data['author'])  # TODO: Update after we can parse in users.
        self.channel_id = snowflake.Snowflake(data['channel_id'])
        if 'components' in data:
            self.components = list()
            for component in data['components']:
                self.components.append(component)
        if 'content' in data:
            self.content = data['content']
        if 'edited_timestamp' in data and data['edited_timestamp'] is not None:
            self.edited_timestamp = datetime.datetime.fromisoformat(data['edited_timestamp'])
        if 'embeds' in data:
            self.embeds = list()
            for embed in data['embeds']:
                self.embeds.append(embed)
        if 'flags' in data:
            self.flags = data['flags']
        if 'mention_everyone' in data:
            self.mention_everyone = data['mention_everyone']
        if 'menion_roles' in data:
            self.mention_roles = list()
            for role in data['mention_roles']:
                # TODO: Roles are just ID's here, we need to parse them into actual role objects.
                # self.mention_roles.append(ext_role.Role().from_dict(role))
                pass
        if 'mentions' in data:
            self.mentions = list()
            for user_dict in data['mentions']:
                self.mentions.append(ext_user.User().from_dict(user_dict))
        if 'pinned' in data:
            self.pinned = data['pinned']
        if 'timestamp' in data:
            self.timestamp = datetime.datetime.fromisoformat(data['timestamp'])
        if 'tts' in data:
            self.tts = data['tts']
        if 'type' in data:
            self.type = enumerations.MESSAGE_TYPE(data['type'])

        # Optional fields
        if 'guild_id' in data:
            self.guild_id = snowflake.Snowflake(data['guild_id'])
        if 'member' in data:
            self.member = ext_user.Member().from_dict(data['member'])  # TODO: Update after we can parse in users.
            if self.author is not None:
                self.member.update_from_user(self.author)
        if 'nonce' in data:
            self.nonce = data['nonce']
        if 'webhook_id' in data:
            self.webhook_id = snowflake.Snowflake(data['webhook_id'])
        if 'message_reference' in data:
            self.message_reference = MessageReference().from_dict(data['message_reference'])
        if 'referenced_message' in data and data['referenced_message'] is not None:
            self.referenced_message = Message().from_dict(data['referenced_message'])
        if 'mention_channels' in data:
            self.mention_channels = list()
            for user_dict in data['mention_channels']:
                self.mention_channels.append(ext_channel.ChannelImporter().from_dict(user_dict))

        return self

    def to_sendable_dict(self) -> dict:
        '''Sending a message only allows a subset of attributes. Ignore anything else about this message when producing that dict.'''
        new_dict: Dict[str, object] = dict()
        if self.content is not None:
            new_dict['content'] = self.content
        new_dict['tts'] = self.tts if (self.tts is not None) else False
        # new_dict['file'] = None  # TODO: Handle a file upload.
        if self.embeds is not None:
            new_dict['embeds'] = list()  # TODO: Handle embeds.
            assert type(new_dict['embeds']) is list
            assert type(self.embeds) is list
            for embed in self.embeds:
                new_dict['embeds'].append(embed.to_dict())
        # new_dict['allowed_mentions'] = True  # BUG: This isn't a bool, its an `AllowedMentionsObject`, which we don't support yet.
        # new_dict['message_reference'] = None
        # new_dict['sticker_ids'] = None

        if isinstance(self.components, list) and len(self.components):
            new_dict['components'] = list()
            assert type(new_dict['components']) is list
            for component in self.components:
                new_dict['components'].append(component.to_dict())

        return new_dict

    def validate(self):
        '''Validate object is prepared for dispatch to discord.'''
        assert (self.content is not None) or (self.embeds is not None) or (self.attachments is not None),\
            'Message must have a content, embeds, or attachments to be valid for sending.'

        if self.content is not None:
            assert type(self.content) is str,\
                f'Got invalid type {type(self.content)} in Message.content, must be str.'

            assert len(self.content) <= 2000,\
                f'Invalid length of {len(self.content):,} in Message.content, must be <= 2000 characters.'

        if self.components is not None:
            assert type(self.components) is list,\
                f'Got invalid type {type(self.components)} in Message.components, must be list.'

            assert len(self.components) <= 5,\
                f'Invalid length of {len(self.components):,} in Message.components, must be <= 5 elements.'

            custom_ids = list()
            for component in self.components:
                assert type(component) is ext_components.ActionRow,\
                    f'Got invalid type {type(component)} in Message.components, must be ActionRow.'
                component.validate()
                for sub_component in component.components:
                    if sub_component.custom_id is not None:
                        assert sub_component.custom_id not in custom_ids,\
                            f'Found duplicate custom_id [{sub_component.custom_id}]'
                        custom_ids.append(sub_component.custom_id)

        if self.embeds is not None:
            assert type(self.embeds) is list,\
                f'Got invalid type {type(self.embeds)} in Message.embeds, must be list.'

            assert len(self.embeds) <= 25,\
                f'Invalid length of {len(self.embeds):,} in Message.embeds, must be <= 5 elements.'

            for embed in self.embeds:
                embed.validate()

    class formatter:
        '''Helper formatting functions.

        These mostly implement the functions found in the [Discord API docs](https://discord.com/developers/docs/reference#message-formatting).


        Examples:
            When generating a string for a message, the formatter class provides a quick shortcut to get and use the various formatters.

            ```python
            message = Message()
            f'The time is {message.formatter.timestamp()}!'
            ```

        '''

        class TIMESTAMP_FLAGS(enum.Enum):
            '''Types of timestamp displays.

            Attributes:
                SHORT_TIME (enum): 16:20
                LONG_TIME (enum): 16:20:30
                SHORT_DATE (enum): 20/04/2021
                SHORT_DATE_TIME (enum): 20 April 2021 16:20
                LONG_DATE (enum): 20 April 2021
                LONG_DATE_TIME (enum): Tuesday, 20 April 2021 16:20
                RELATIVE_TIME (enum): 2 months ago
                DEFAULT (enum): 20 April 2021 16:20
            '''
            SHORT_TIME = 't'
            LONG_TIME = 'T'
            SHORT_DATE = 'd'
            LONG_DATE = 'D'
            SHORT_DATE_TIME = 'f'
            LONG_DATE_TIME = 'F'
            RELATIVE_TIME = 'R'

            DEFAULT = SHORT_DATE_TIME

        @classmethod
        def timestamp(cls, timestamp: datetime.datetime = None, flag: Union[TIMESTAMP_FLAGS, str] = TIMESTAMP_FLAGS.DEFAULT) -> str:
            '''Return valid timestamp string. Discord will display this in a useful way.

            Arguments:
                timestamp (datetime): The timestamp to parse, can be omitted to use `now`.
                flag (TIMESTAMP_FLAGS|str): Flag to use when parsing the timestamp on the client.

            Returns:
                str: Valid string to use as part of a discord message to reference to a point in time.
            '''
            if timestamp is None:
                timestamp = datetime.datetime.now()
            if type(flag) is cls.TIMESTAMP_FLAGS:
                flag = flag.value
            return f'<t:{int(timestamp.timestamp())}:{flag}>'

        @classmethod
        def user(cls, user: 'Union[snowflake.Snowflake, ext_user.User, ext_user.Member]'):
            '''Return valid mention string for the user's name.

            Arguments:
                user (Snowflake|User|Member): User or User's ID to mention.

            Returns:
                str: Valid string to use as part of a discord message to reference a User by name.
            '''
            if isinstance(user, (snowflake.Snowflake, str)):
                user_id = user
            elif isinstance(user, (ext_user.User, ext_user.Member)):
                user_id = user.id
            return f'<@{user_id}>'

        @classmethod
        def user_nickname(cls, user: 'Union[snowflake.Snowflake, ext_user.User, ext_user.Member]'):
            '''Return valid mention string for the user's nickname.

            Arguments:
                user (Snowflake|User|Member): User or User's ID to mention.

            Returns:
                str: Valid string to use as part of a discord message to reference a User by nickname.
            '''
            if isinstance(user, (snowflake.Snowflake, str)):
                user_id = user
            elif isinstance(user, (ext_user.User, ext_user.Member)):
                user_id = user.id
            return f'<@!{user_id}>'

        @classmethod
        def channel(cls, channel: 'Union[snowflake.Snowflake, ext_channel.Channel]'):
            '''Return valid mention string for the channel.

            Arguments:
                channel (Snowflake|Channel): Channel or channel ID to mention. Any subclass of Channel is accepted.

            Returns:
                str: Valid string to use as part of a discord message to reference a Channel.
            '''
            if isinstance(channel, (snowflake.Snowflake, str)):
                channel_id = channel
            elif isinstance(channel, ext_channel.Channel):
                channel_id = channel.id
            return f'<#{channel_id}>'

        @classmethod
        def role(cls, role: 'Union[snowflake.Snowflake, ext_role.Role]'):
            '''Return valid mention string for the role.

            Arguments:
                role (Snowflake|Role): Role or role's ID to mention.

            Returns:
                str: Valid string to use as part of a discord message to reference a Role.
            '''
            if isinstance(role, (snowflake.Snowflake, str)):
                role_id = role
            elif isinstance(role, ext_role.Role):
                role_id = role.id
            return f'<@&{role_id}>'


class MessageReference(BaseDiscordObject):
    '''Slim data holding class.

    Attributes:
        message_id (Snowflake|None): ID of the originating message.
        channel_id: (Snowflake|None): ID of the originating message's channel.
        guild_id: (Snowflake|None): ID of the originating message's guild.
        fail_if_not_exists (bool|None): When sending, whether to error if the referenced message doesn't exist instead of sending as a normal (non-reply) message, default true

    '''

    message_id: Optional[snowflake.Snowflake] = None  # id of the originating message
    channel_id: Optional[snowflake.Snowflake] = None  # id of the originating message's channel
    guild_id: Optional[snowflake.Snowflake] = None  # id of the originating message's guild
    fail_if_not_exists: Optional[bool] = None  # when sending, whether to error if the referenced message doesn't exist instead of sending as a normal (non-reply) message, default true

    def from_dict(self, data: dict) -> 'MessageReference':
        '''Parse a MessageReference from an API compliant dict.'''
        if 'message_id' in data:
            self.message_id = snowflake.Snowflake(data['message_id'])
        if 'channel_id' in data:
            self.channel_id = snowflake.Snowflake(data['channel_id'])
        if 'guild_id' in data:
            self.guild_id = snowflake.Snowflake(data['guild_id'])
        if 'fail_if_not_exists' in data:
            self.fail_if_not_exists = data['fail_if_not_exists']

        return self


class MessageUpdate(Message):
    '''Duplicate of the Message class, but most fields are now annotated as optional.'''
    guild_id: 'Optional[snowflake.Snowflake]'  # type: ignore
    author: 'Optional[ext_user.User]'  # type: ignore
    member: 'Optional[ext_user.Member]'  # type: ignore
    content: Optional[str]  # type: ignore
    timestamp: Optional[datetime.datetime]  # type: ignore
    edited_timestamp: Optional[datetime.datetime]  # type: ignore
    tts: Optional[bool]  # type: ignore
    mention_everyone: Optional[bool]  # type: ignore
    mentions: 'Optional[List[ext_user.User]]'  # type: ignore
    mention_roles: 'Optional[List[ext_role.Role]]'  # type: ignore
    mention_channels: 'Optional[List[ext_channel.Channel]]'  # type: ignore
    attachments: Optional[List]  # type: ignore
    embeds: Optional[List[ext_embed.Embed]]  # type: ignore
    # reactions: 'List[reaction.Reaction]'   # type: ignore
    nonce: Optional[Union[int, str]]  # type: ignore
    pinned: Optional[bool]  # type: ignore
    webhook_id: Optional['snowflake.Snowflake']  # type: ignore
    type: 'Optional[enumerations.MESSAGE_TYPE]'  # type: ignore
    # activity  # Activity TBD  # type: ignore
    # application: application.Application  # Application TBD  # type: ignore
    application_id: 'Optional[snowflake.Snowflake]'  # type: ignore
    message_reference: 'Optional[MessageReference]'  # type: ignore
    flags: Optional[int]  # type: ignore
    referenced_message: 'Optional[Message]'  # type: ignore
    # interaction  # type: ignore
    thread: 'Optional[ext_channel.Channel]'  # type: ignore
    components: 'Optional[List[ext_components.Component]]'  # type: ignore
    # sticker_items: Optional[List[Sticker]] = []  # type: ignore



from typing import Optional

from ..objects.base_object import BaseDiscordObject
from .. import objects


async def on_any(self, object: Optional[BaseDiscordObject], raw_event: dict):
    '''Called for every event.

    Arguments:
        self (DiscordClient): Client
        object (BaseDiscordObject): Python object representing the event.
        raw_event (dict): Raw dict from discord API.
    '''
    pass


async def on_channel_create(self, object: BaseDiscordObject, raw_object: dict):
    '''Empty placeholder for given event.

    Arguments:
        self (DiscordClient): Client
        object (BaseDiscordObject): Python object representing the event.
        raw_object (dict): Raw dict from discord API.

    '''
    pass


async def on_channel_delete(self, object: BaseDiscordObject, raw_object: dict):
    '''Empty placeholder for given event.

    Arguments:
        self (DiscordClient): Client
        object (BaseDiscordObject): Python object representing the event.
        raw_object (dict): Raw dict from discord API.

    '''
    pass


async def on_channel_pins_update(self, object: BaseDiscordObject, raw_object: dict):
    '''Empty placeholder for given event.

    Arguments:
        self (DiscordClient): Client
        object (BaseDiscordObject): Python object representing the event.
        raw_object (dict): Raw dict from discord API.

    '''
    pass


async def on_channel_update(self, object: BaseDiscordObject, raw_object: dict):
    '''Empty placeholder for given event.

    Arguments:
        self (DiscordClient): Client
        object (BaseDiscordObject): Python object representing the event.
        raw_object (dict): Raw dict from discord API.

    '''
    pass


async def on_guild_ban_add(self, object: BaseDiscordObject, raw_object: dict):
    '''Empty placeholder for given event.

    Arguments:
        self (DiscordClient): Client
        object (BaseDiscordObject): Python object representing the event.
        raw_object (dict): Raw dict from discord API.

    '''
    pass


async def on_guild_ban_remove(self, object: BaseDiscordObject, raw_object: dict):
    '''Empty placeholder for given event.

    Arguments:
        self (DiscordClient): Client
        object (BaseDiscordObject): Python object representing the event.
        raw_object (dict): Raw dict from discord API.

    '''
    pass


async def on_guild_create(self, object: BaseDiscordObject, raw_object: dict):
    '''Empty placeholder for given event.

    Arguments:
        self (DiscordClient): Client
        object (BaseDiscordObject): Python object representing the event.
        raw_object (dict): Raw dict from discord API.

    '''
    pass


async def on_guild_delete(self, object: BaseDiscordObject, raw_object: dict):
    '''Empty placeholder for given event.

    Arguments:
        self (DiscordClient): Client
        object (BaseDiscordObject): Python object representing the event.
        raw_object (dict): Raw dict from discord API.

    '''
    pass


async def on_guild_emojis_update(self, object: BaseDiscordObject, raw_object: dict):
    '''Empty placeholder for given event.

    Arguments:
        self (DiscordClient): Client
        object (BaseDiscordObject): Python object representing the event.
        raw_object (dict): Raw dict from discord API.

    '''
    pass


async def on_guild_integrations_update(self, object: BaseDiscordObject, raw_object: dict):
    '''Empty placeholder for given event.

    Arguments:
        self (DiscordClient): Client
        object (BaseDiscordObject): Python object representing the event.
        raw_object (dict): Raw dict from discord API.

    '''
    pass


async def on_guild_member_add(self, object: BaseDiscordObject, raw_object: dict):
    '''Empty placeholder for given event.

    Arguments:
        self (DiscordClient): Client
        object (BaseDiscordObject): Python object representing the event.
        raw_object (dict): Raw dict from discord API.

    '''
    pass


async def on_guild_member_remove(self, object: BaseDiscordObject, raw_object: dict):
    '''Empty placeholder for given event.

    Arguments:
        self (DiscordClient): Client
        object (BaseDiscordObject): Python object representing the event.
        raw_object (dict): Raw dict from discord API.

    '''
    pass


async def on_guild_member_update(self, object: BaseDiscordObject, raw_object: dict):
    '''Empty placeholder for given event.

    Arguments:
        self (DiscordClient): Client
        object (BaseDiscordObject): Python object representing the event.
        raw_object (dict): Raw dict from discord API.

    '''
    pass


async def on_guild_role_create(self, object: BaseDiscordObject, raw_object: dict):
    '''Empty placeholder for given event.

    Arguments:
        self (DiscordClient): Client
        object (BaseDiscordObject): Python object representing the event.
        raw_object (dict): Raw dict from discord API.

    '''
    pass


async def on_guild_role_delete(self, object: BaseDiscordObject, raw_object: dict):
    '''Empty placeholder for given event.

    Arguments:
        self (DiscordClient): Client
        object (BaseDiscordObject): Python object representing the event.
        raw_object (dict): Raw dict from discord API.

    '''
    pass


async def on_guild_role_update(self, object: BaseDiscordObject, raw_object: dict):
    '''Empty placeholder for given event.

    Arguments:
        self (DiscordClient): Client
        object (BaseDiscordObject): Python object representing the event.
        raw_object (dict): Raw dict from discord API.

    '''
    pass


async def on_guild_stickers_update(self, object: BaseDiscordObject, raw_object: dict):
    '''Empty placeholder for given event.

    Arguments:
        self (DiscordClient): Client
        object (BaseDiscordObject): Python object representing the event.
        raw_object (dict): Raw dict from discord API.

    '''
    pass


async def on_guild_update(self, object: BaseDiscordObject, raw_object: dict):
    '''Empty placeholder for given event.

    Arguments:
        self (DiscordClient): Client
        object (BaseDiscordObject): Python object representing the event.
        raw_object (dict): Raw dict from discord API.

    '''
    pass


async def on_integration_create(self, object: BaseDiscordObject, raw_object: dict):
    '''Empty placeholder for given event.

    Arguments:
        self (DiscordClient): Client
        object (BaseDiscordObject): Python object representing the event.
        raw_object (dict): Raw dict from discord API.

    '''
    pass


async def on_integration_delete(self, object: BaseDiscordObject, raw_object: dict):
    '''Empty placeholder for given event.

    Arguments:
        self (DiscordClient): Client
        object (BaseDiscordObject): Python object representing the event.
        raw_object (dict): Raw dict from discord API.

    '''
    pass


async def on_integration_update(self, object: BaseDiscordObject, raw_object: dict):
    '''Empty placeholder for given event.

    Arguments:
        self (DiscordClient): Client
        object (BaseDiscordObject): Python object representing the event.
        raw_object (dict): Raw dict from discord API.

    '''
    pass


async def on_invite_create(self, object: BaseDiscordObject, raw_object: dict):
    '''Empty placeholder for given event.

    Arguments:
        self (DiscordClient): Client
        object (BaseDiscordObject): Python object representing the event.
        raw_object (dict): Raw dict from discord API.

    '''
    pass


async def on_invite_delete(self, object: BaseDiscordObject, raw_object: dict):
    '''Empty placeholder for given event.

    Arguments:
        self (DiscordClient): Client
        object (BaseDiscordObject): Python object representing the event.
        raw_object (dict): Raw dict from discord API.

    '''
    pass


async def on_message_create(self, message: objects.Message, raw_message: dict):
    '''Empty placeholder for given event.

    Arguments:
        self (DiscordClient): Client
        message (Message): Python object representing the event.
        raw_message (dict): Raw dict from discord API.

    '''
    pass


async def on_message_delete(self, object: BaseDiscordObject, raw_object: dict):
    '''Empty placeholder for given event.

    Arguments:
        self (DiscordClient): Client
        object (BaseDiscordObject): Python object representing the event.
        raw_object (dict): Raw dict from discord API.

    '''
    pass


async def on_message_delete_bulk(self, object: BaseDiscordObject, raw_object: dict):
    '''Empty placeholder for given event.

    Arguments:
        self (DiscordClient): Client
        object (BaseDiscordObject): Python object representing the event.
        raw_object (dict): Raw dict from discord API.

    '''
    pass


async def on_message_reaction_add(self, object: BaseDiscordObject, raw_object: dict):
    '''Empty placeholder for given event.

    Arguments:
        self (DiscordClient): Client
        object (BaseDiscordObject): Python object representing the event.
        raw_object (dict): Raw dict from discord API.

    '''
    pass


async def on_message_reaction_remove(self, object: BaseDiscordObject, raw_object: dict):
    '''Empty placeholder for given event.

    Arguments:
        self (DiscordClient): Client
        object (BaseDiscordObject): Python object representing the event.
        raw_object (dict): Raw dict from discord API.

    '''
    pass


async def on_message_reaction_remove_all(self, object: BaseDiscordObject, raw_object: dict):
    '''Empty placeholder for given event.

    Arguments:
        self (DiscordClient): Client
        object (BaseDiscordObject): Python object representing the event.
        raw_object (dict): Raw dict from discord API.

    '''
    pass


async def on_message_reaction_remove_emoji(self, object: BaseDiscordObject, raw_object: dict):
    '''Empty placeholder for given event.

    Arguments:
        self (DiscordClient): Client
        object (BaseDiscordObject): Python object representing the event.
        raw_object (dict): Raw dict from discord API.

    '''
    pass


async def on_message_update(self, message: objects.MessageUpdate, raw_object: dict):
    '''Empty placeholder for given event.

    Arguments:
        self (DiscordClient): Client
        message (MessageUpdate): Updated message. Note that many fields may be missing from this entity!
        raw_object (dict): Raw dict from discord API.

    '''
    pass


async def on_presence_update(self, object: BaseDiscordObject, raw_object: dict):
    '''Empty placeholder for given event.

    Arguments:
        self (DiscordClient): Client
        object (BaseDiscordObject): Python object representing the event.
        raw_object (dict): Raw dict from discord API.

    '''
    pass


async def on_ready(self, ready: objects.Ready, raw_ready: dict):
    '''Empty placeholder for given event.

    Arguments:
        self (DiscordClient): Client
        ready (Ready): Python object representing the event.
        raw_ready (dict): Raw dict from discord API.

    '''
    pass


async def on_stage_instance_create(self, object: BaseDiscordObject, raw_object: dict):
    '''Empty placeholder for given event.

    Arguments:
        self (DiscordClient): Client
        object (BaseDiscordObject): Python object representing the event.
        raw_object (dict): Raw dict from discord API.

    '''
    pass


async def on_stage_instance_delete(self, object: BaseDiscordObject, raw_object: dict):
    '''Empty placeholder for given event.

    Arguments:
        self (DiscordClient): Client
        object (BaseDiscordObject): Python object representing the event.
        raw_object (dict): Raw dict from discord API.

    '''
    pass


async def on_stage_instance_update(self, object: BaseDiscordObject, raw_object: dict):
    '''Empty placeholder for given event.

    Arguments:
        self (DiscordClient): Client
        object (BaseDiscordObject): Python object representing the event.
        raw_object (dict): Raw dict from discord API.

    '''
    pass


async def on_thread_create(self, object: BaseDiscordObject, raw_object: dict):
    '''Empty placeholder for given event.

    Arguments:
        self (DiscordClient): Client
        object (BaseDiscordObject): Python object representing the event.
        raw_object (dict): Raw dict from discord API.

    '''
    pass


async def on_thread_delete(self, object: BaseDiscordObject, raw_object: dict):
    '''Empty placeholder for given event.

    Arguments:
        self (DiscordClient): Client
        object (BaseDiscordObject): Python object representing the event.
        raw_object (dict): Raw dict from discord API.

    '''
    pass


async def on_thread_list_sync(self, object: BaseDiscordObject, raw_object: dict):
    '''Empty placeholder for given event.

    Arguments:
        self (DiscordClient): Client
        object (BaseDiscordObject): Python object representing the event.
        raw_object (dict): Raw dict from discord API.

    '''
    pass


async def on_thread_member_update(self, object: BaseDiscordObject, raw_object: dict):
    '''Empty placeholder for given event.

    Arguments:
        self (DiscordClient): Client
        object (BaseDiscordObject): Python object representing the event.
        raw_object (dict): Raw dict from discord API.

    '''
    pass


async def on_thread_members_update(self, object: BaseDiscordObject, raw_object: dict):
    '''Empty placeholder for given event.

    Arguments:
        self (DiscordClient): Client
        object (BaseDiscordObject): Python object representing the event.
        raw_object (dict): Raw dict from discord API.

    '''
    pass


async def on_thread_update(self, object: BaseDiscordObject, raw_object: dict):
    '''Empty placeholder for given event.

    Arguments:
        self (DiscordClient): Client
        object (BaseDiscordObject): Python object representing the event.
        raw_object (dict): Raw dict from discord API.

    '''
    pass


async def on_typing_start(self, object: BaseDiscordObject, raw_object: dict):
    '''Empty placeholder for given event.

    Arguments:
        self (DiscordClient): Client
        object (BaseDiscordObject): Python object representing the event.
        raw_object (dict): Raw dict from discord API.

    '''
    pass


async def on_voice_state_update(self, object: BaseDiscordObject, raw_object: dict):
    '''Empty placeholder for given event.

    Arguments:
        self (DiscordClient): Client
        object (BaseDiscordObject): Python object representing the event.
        raw_object (dict): Raw dict from discord API.

    '''
    pass


async def on_webhooks_update(self, object: BaseDiscordObject, raw_object: dict):
    '''Empty placeholder for given event.

    Arguments:
        self (DiscordClient): Client
        object (BaseDiscordObject): Python object representing the event.
        raw_object (dict): Raw dict from discord API.

    '''
    pass


async def on_interaction_create(self, object: BaseDiscordObject, raw_object: dict):
    '''Empty placeholder for given event.

    Arguments:
        self (DiscordClient): Client
        object (BaseDiscordObject): Python object representing the event.
        raw_object (dict): Raw dict from discord API.

    '''
    pass
